from __future__ import unicode_literals

import logging
import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers, exceptions

from waldur_core.core import serializers as core_serializers
from waldur_core.media.serializers import ProtectedMediaSerializerMixin
from waldur_core.structure import models as structure_models, SupportedServices, serializers as structure_serializers
from waldur_jira import serializers as jira_serializers
from waldur_mastermind.common.serializers import validate_options
from waldur_mastermind.support.backend.atlassian import ServiceDeskBackend

from . import models

logger = logging.getLogger(__name__)
User = get_user_model()


def render_issue_template(config_name, issue):
    issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})
    if not issue_settings:
        return ''

    raw = issue_settings[config_name]
    template = Template(raw)
    return template.render(Context({'issue': issue}, autoescape=False))


class IssueSerializer(core_serializers.AugmentedSerializerMixin,
                      serializers.HyperlinkedModelSerializer):
    resource = core_serializers.GenericRelatedField(
        related_models=structure_models.ResourceMixin.get_all_models(), required=False)
    caller = serializers.HyperlinkedRelatedField(
        view_name='user-detail',
        lookup_field='uuid',
        queryset=User.objects.all(),
        required=False,
        allow_null=True,
    )
    reporter = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        read_only=True
    )
    assignee = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        queryset=models.SupportUser.objects.all(),
        required=False,
        allow_null=True,
    )
    template = serializers.HyperlinkedRelatedField(
        view_name='support-template-detail',
        lookup_field='uuid',
        queryset=models.Template.objects.all(),
        required=False,
        allow_null=True,
    )
    resource_type = serializers.SerializerMethodField()
    resource_name = serializers.ReadOnlyField(source='resource.name')
    type = serializers.ChoiceField(
        choices=[(t, t) for t in settings.WALDUR_SUPPORT['ISSUE']['types']],
        initial=settings.WALDUR_SUPPORT['ISSUE']['types'][0],
        default=settings.WALDUR_SUPPORT['ISSUE']['types'][0])
    is_reported_manually = serializers.BooleanField(
        initial=False, default=False, write_only=True,
        help_text=_('Set true if issue is created by regular user via portal.'))

    class Meta(object):
        model = models.Issue
        fields = (
            'url', 'uuid', 'type', 'key', 'backend_id', 'link',
            'summary', 'description', 'status', 'resolution', 'priority',
            'caller', 'caller_uuid', 'caller_full_name',
            'reporter', 'reporter_uuid', 'reporter_name',
            'assignee', 'assignee_uuid', 'assignee_name',
            'customer', 'customer_uuid', 'customer_name',
            'project', 'project_uuid', 'project_name',
            'resource', 'resource_type', 'resource_name',
            'created', 'modified', 'is_reported_manually',
            'first_response_sla', 'template'
        )
        read_only_fields = ('key', 'status', 'resolution', 'backend_id', 'link', 'first_response_sla')
        protected_fields = ('customer', 'project', 'resource', 'type', 'caller', 'template', 'priority')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            customer={'lookup_field': 'uuid', 'view_name': 'customer-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
        )
        related_paths = dict(
            caller=('uuid', 'full_name',),
            reporter=('uuid', 'name',),
            assignee=('uuid', 'name',),
            customer=('uuid', 'name',),
            project=('uuid', 'name',),
        )

    def get_fields(self):
        fields = super(IssueSerializer, self).get_fields()

        if 'view' not in self.context:  # On docs generation context does not contain "view".
            return fields

        user = self.context['view'].request.user
        if not user.is_staff and not user.is_support:
            del fields['link']

        return fields

    def get_resource_type(self, obj):
        if isinstance(obj.resource, structure_models.ResourceMixin):
            return SupportedServices.get_name_for_model(obj.resource_content_type.model_class())

    def validate(self, attrs):
        if self.instance is not None:
            return attrs
        if attrs.pop('is_reported_manually'):
            attrs['caller'] = self.context['request'].user
            if attrs.get('assignee'):
                raise serializers.ValidationError(
                    {'assignee': _('Assignee cannot be defined if issue is reported manually.')})
        else:
            if not attrs.get('caller'):
                raise serializers.ValidationError({'caller': _('This field is required.')})
            reporter = models.SupportUser.objects.filter(
                user=self.context['request'].user, is_active=True).first()
            if not reporter:
                raise serializers.ValidationError(
                    _('You cannot report issues because your help desk account is not connected to profile.'))
            attrs['reporter'] = reporter
        return attrs

    def validate_customer(self, customer):
        """ User has to be customer owner, staff or global support """
        if not customer:
            return customer
        user = self.context['request'].user
        if (not customer or
                user.is_staff or
                user.is_support or
                customer.has_user(user, structure_models.CustomerRole.OWNER)):
            return customer
        raise serializers.ValidationError(_('Only customer owner, staff or support can report customer issues.'))

    def validate_project(self, project):
        if not project:
            return project
        user = self.context['request'].user
        if (not project or
                user.is_staff or
                user.is_support or
                project.customer.has_user(user, structure_models.CustomerRole.OWNER) or
                project.has_user(user, structure_models.ProjectRole.MANAGER) or
                project.has_user(user, structure_models.ProjectRole.ADMINISTRATOR)):
            return project
        raise serializers.ValidationError(
            _('Only customer owner, project manager, project admin, staff or support can report such issue.'))

    def validate_resource(self, resource):
        if resource:
            self.validate_project(resource.service_project_link.project)
        return resource

    def validate_priority(self, priority):
        user = self.context['request'].user
        if not user.is_staff and not user.is_support:
            raise serializers.ValidationError(_('Only staff or support can specify issue priority.'))
        try:
            models.Priority.objects.get(name=priority)
        except (models.Priority.DoesNotExist, models.Priority.MultipleObjectsReturned):
            raise serializers.ValidationError(_('Priority with requested name does not exist.'))
        return priority

    @transaction.atomic()
    def create(self, validated_data):
        resource = validated_data.get('resource')
        if resource:
            validated_data['project'] = resource.service_project_link.project
        project = validated_data.get('project')
        if project:
            validated_data['customer'] = project.customer

        validated_data['description'] = render_issue_template('description', validated_data)
        validated_data['summary'] = render_issue_template('summary', validated_data)
        return super(IssueSerializer, self).create(validated_data)

    def _render_template(self, config_name, issue):
        raw = self.issue_settings[config_name]
        template = Template(raw)
        return template.render(Context({'issue': issue}))


class PrioritySerializer(core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Priority
        fields = ('url', 'uuid', 'name', 'description', 'icon_url')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class CommentSerializer(core_serializers.AugmentedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    # should be initialized with issue in context on creation
    author_user = serializers.HyperlinkedRelatedField(
        source='author.user',
        view_name='user-detail',
        lookup_field='uuid',
        read_only=True,
    )

    author_uuid = serializers.ReadOnlyField(source='author.user.uuid')

    class Meta(object):
        model = models.Comment
        fields = ('url', 'uuid', 'issue', 'issue_key', 'description', 'is_public',
                  'author_name', 'author_uuid', 'author_user', 'backend_id', 'created')
        read_only_fields = ('issue', 'backend_id',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        )
        related_paths = dict(
            author=('name',),
            issue=('key',),
        )

    @transaction.atomic()
    def create(self, validated_data):
        author_user = self.context['request'].user
        validated_data['author'], _ = models.SupportUser.objects.get_or_create_from_user(author_user)
        validated_data['issue'] = self.context['view'].get_object()
        return super(CommentSerializer, self).create(validated_data)


class SupportUserSerializer(core_serializers.AugmentedSerializerMixin,
                            serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.SupportUser
        fields = ('url', 'uuid', 'name', 'backend_id', 'user')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            user={'lookup_field': 'uuid', 'view_name': 'user-detail'}
        )


class WebHookReceiverSerializer(jira_serializers.WebHookReceiverSerializer):
    def get_project(self, project_key):
        class Project:
            def get_backend(self):
                return ServiceDeskBackend()

        return Project()

    def get_issue(self, project, key, create):
        issue = None

        try:
            issue = models.Issue.objects.get(backend_id=key)
        except models.Issue.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Issue with id %s does not exist.' % key)

        return issue

    def get_comment(self, issue, key, create):
        comment = None

        try:
            comment = models.Comment.objects.get(issue=issue, backend_id=key)
        except models.Comment.DoesNotExist:
            if not create:
                raise serializers.ValidationError('Comment with id %s does not exist.' % key)

        return comment


WebHookReceiverSerializer.remove_event(['jira:issue_created'])


class OfferingSerializer(structure_serializers.PermissionFieldFilteringMixin,
                         core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    type = serializers.ReadOnlyField(source='template.name')

    template = serializers.HyperlinkedRelatedField(
        queryset=models.OfferingTemplate.objects.all(),
        view_name='support-offering-template-detail',
        lookup_field='uuid',
    )
    plan = serializers.HyperlinkedRelatedField(
        queryset=models.OfferingPlan.objects.all(),
        view_name='support-offering-plan-detail',
        lookup_field='uuid',
        required=False,
        write_only=True,
    )
    state = serializers.ReadOnlyField(source='get_state_display')
    report = serializers.JSONField(required=False)
    template_uuid = serializers.ReadOnlyField(source='template.uuid')
    resource_type = serializers.SerializerMethodField()

    class Meta(object):
        model = models.Offering
        fields = ('url', 'uuid', 'name', 'project', 'type', 'template', 'template_uuid', 'resource_type', 'plan',
                  'state', 'type_label', 'unit_price',
                  'unit', 'created', 'modified', 'issue', 'issue_name', 'issue_link',
                  'issue_key', 'issue_description', 'issue_uuid', 'issue_status',
                  'project_name', 'project_uuid', 'product_code', 'article_code', 'report')
        read_only_fields = ('type_label', 'issue', 'unit_price', 'unit', 'state', 'product_code', 'article_code')
        protected_fields = ('project', 'type', 'template', 'plan')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
            unit_price={'decimal_places': 2},
        )
        related_paths = dict(
            issue=('uuid', 'name', 'status', 'key', 'description', 'link'),
            project=('uuid', 'name',),
        )

    def validate_report(self, report):
        if not isinstance(report, list):
            raise serializers.ValidationError('Report should be a list.')

        if len(report) == 0:
            raise serializers.ValidationError('Report object should contain at least one section.')

        for section in report:
            if not isinstance(section, dict):
                raise serializers.ValidationError('Report section should be an object.')

            if not section.get('header'):
                raise serializers.ValidationError('Report section should contain header.')

            if not section.get('body'):
                raise serializers.ValidationError('Report section should contain body.')

        return report

    def get_filtered_field_names(self):
        return ('project',)

    def get_resource_type(self, obj):
        return obj.get_scope_type()


class ConfigurableFormDescriptionMixin(object):
    def _form_description(self, configuration, validated_data):
        result = []

        for key in configuration.get('order') or []:
            if key not in validated_data:
                continue

            label = configuration['options'].get(key, {})
            label_value = label.get('label', key)
            result.append('%s: \'%s\'' % (label_value, validated_data[key]))

        if 'description' in validated_data:
            result.append('\n %s' % validated_data['description'])

        return '\n'.join(result)


class OfferingCreateSerializer(OfferingSerializer, ConfigurableFormDescriptionMixin):
    attributes = serializers.JSONField(required=False, write_only=True, allow_null=True)
    description = serializers.CharField(required=False, help_text=_('Description to add to the issue.'))

    class Meta(OfferingSerializer.Meta):
        fields = OfferingSerializer.Meta.fields + ('description', 'attributes')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid',
                     'view_name': 'project-detail',
                     'required': True,
                     'allow_empty': False,
                     'allow_null': False},
        )

    def create(self, validated_data):
        """
        Each offering corresponds to the single issue which has next values:
            'project' - a hyperlinked field which must be provided with every request;
            'customer' - customer is extracted from the provided project;
            'caller' - a user who sent a request is considered to be a 'caller' of the issue;
            'summary' - has a format of 'Request for <template_name>' or 'Request for "Support" if empty;
            'description' - combined list of all other fields provided with the request;
        """
        project = validated_data['project']
        template = validated_data['template']
        plan = validated_data.pop('plan', None)

        # Temporary code for backward compatibility
        if not plan:
            plan = template.plans.first()

        # It's okay if there's no plan.
        if plan and plan.template != template:
            raise serializers.ValidationError({
                'plan': _('Plan should be related to the same template.'),
            })

        attributes = validated_data.get('attributes', {})
        if isinstance(template.config, dict) and template.config.get('options'):
            try:
                validate_options(template.config['options'], attributes)
            except serializers.ValidationError as exc:
                raise serializers.ValidationError({'attributes': exc})
            else:
                validated_data.update(attributes)

        offering_configuration = template.config
        type_label = offering_configuration.get('label', template.name)
        issue_details = dict(
            caller=self.context['request'].user,
            project=project,
            customer=project.customer,
            type=settings.WALDUR_SUPPORT['DEFAULT_OFFERING_ISSUE_TYPE'],
            summary='Request for \'%s\'' % type_label,
            description=self._form_description(offering_configuration, validated_data))
        issue_details['summary'] = render_issue_template('summary', issue_details)
        issue_details['description'] = render_issue_template('description', issue_details)
        issue = models.Issue.objects.create(**issue_details)

        payload = dict(
            issue=issue,
            project=issue.project,
            name=validated_data.get('name'),
            template=template,
        )
        if plan:
            payload.update(dict(
                product_code=plan.product_code,
                article_code=plan.article_code,
                unit_price=plan.unit_price,
                unit=plan.unit,
                plan=plan,
            ))
        else:
            # Temporary workaround for backward compatibility
            payload.update(dict(
                product_code=offering_configuration.get('product_code', ''),
                article_code=offering_configuration.get('article_code', ''),
            ))

        offering = models.Offering.objects.create(**payload)

        return offering


class OfferingCompleteSerializer(serializers.Serializer):
    unit_price = serializers.DecimalField(max_digits=13, decimal_places=7)
    unit = serializers.ChoiceField(choices=models.Offering.Units.CHOICES, default=models.Offering.Units.PER_DAY)

    def update(self, instance, validated_data):
        instance.unit_price = validated_data['unit_price']
        instance.unit = validated_data['unit']
        instance.state = models.Offering.States.OK
        instance.save(update_fields=['state', 'unit_price', 'unit'])
        return instance


class AttachmentSerializer(ProtectedMediaSerializerMixin,
                           core_serializers.RestrictedSerializerMixin,
                           core_serializers.AugmentedSerializerMixin,
                           serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Attachment
        fields = ('url', 'uuid', 'issue', 'issue_key', 'created', 'file',
                  'mime_type', 'file_size', 'thumbnail', 'backend_id', )
        read_only_fields = ('backend_id',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        )
        related_paths = dict(
            issue=('key',),
        )

    def validate(self, attrs):
        filename, file_extension = os.path.splitext(attrs['file'].name)
        if file_extension in settings.WALDUR_SUPPORT['EXCLUDED_ATTACHMENT_TYPES']:
            raise serializers.ValidationError(_('Invalid file extension'))

        user = self.context['request'].user
        issue = attrs['issue']
        if user.is_staff or \
                (issue.customer and issue.customer.has_user(user, structure_models.CustomerRole.OWNER)) or \
                issue.caller == user:
            return attrs

        raise exceptions.PermissionDenied()


class TemplateAttachmentSerializer(ProtectedMediaSerializerMixin, serializers.ModelSerializer):
    class Meta(object):
        model = models.TemplateAttachment
        fields = ('name', 'file')


class TemplateSerializer(serializers.HyperlinkedModelSerializer):
    attachments = TemplateAttachmentSerializer(many=True)

    class Meta(object):
        model = models.Template
        fields = ('url', 'uuid',
                  'name', 'native_name',
                  'description', 'native_description',
                  'issue_type', 'attachments')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-template-detail'},
        )

    def get_fields(self):
        fields = super(TemplateSerializer, self).get_fields()
        if not settings.WALDUR_CORE['NATIVE_NAME_ENABLED']:
            del fields['native_name']
            del fields['native_description']
        return fields


class OfferingTemplateSerializer(serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.OfferingTemplate
        fields = ('url', 'uuid', 'name', 'config')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-template-detail'},
        )


class OfferingPlanSerializer(serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.OfferingPlan
        fields = ('url', 'uuid', 'product_code', 'article_code', 'unit', 'unit_price')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-plan-detail'},
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.logging.loggers
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0037_remove_customer_billing_backend_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='OpenStackTenantService',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(verbose_name='organization', to='structure.Customer')),
            ],
            options={
                'verbose_name': 'OpenStackTenant service',
                'verbose_name_plural': 'OpenStackTenan services',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OpenStackTenantServiceProjectLink',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('project', models.ForeignKey(to='structure.Project')),
                ('service', models.ForeignKey(to='openstack_tenant.OpenStackTenantService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'OpenStackTenant provider project link',
                'verbose_name_plural': 'OpenStackTenant provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='openstacktenantservice',
            name='projects',
            field=models.ManyToManyField(related_name='openstack_tenant_services', through='openstack_tenant.OpenStackTenantServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='openstacktenantservice',
            name='settings',
            field=models.ForeignKey(to='structure.ServiceSettings'),
        ),
        migrations.AlterUniqueTogether(
            name='openstacktenantserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='openstacktenantservice',
            unique_together=set([('customer', 'settings')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0021_volume_instance'),
    ]

    operations = [
        migrations.AddField(
            model_name='volume',
            name='device',
            field=models.CharField(blank=True, max_length=50, help_text='Name of volume as instance device e.g. /dev/vdb.', validators=[django.core.validators.RegexValidator('^/dev/[a-zA-Z0-9]+$', message='Device should match pattern "/dev/alphanumeric+"')]),
        ),
    ]



from __future__ import unicode_literals

from django.conf.urls import url

from waldur_mastermind.support import views


def register_in(router):
    router.register(r'support-issues', views.IssueViewSet, base_name='support-issue')
    router.register(r'support-priorities', views.PriorityViewSet, base_name='support-priority')
    router.register(r'support-comments', views.CommentViewSet, base_name='support-comment')
    router.register(r'support-users', views.SupportUserViewSet, base_name='support-user')
    router.register(r'support-offerings', views.OfferingViewSet, base_name='support-offering')
    router.register(r'support-attachments', views.AttachmentViewSet, base_name='support-attachment')
    router.register(r'support-templates', views.TemplateViewSet, base_name='support-template')
    router.register(r'support-offering-templates', views.OfferingTemplateViewSet, base_name='support-offering-template')
    router.register(r'support-offering-plans', views.OfferingPlanViewSet, base_name='support-offering-plan')


urlpatterns = [
    url(r'^api/support-jira-webhook/$', views.WebHookReceiverView.as_view(), name='web-hook-receiver'),
]



import datetime

from django.utils.dateparse import parse_datetime
import logging

from waldur_mastermind.marketplace import models as marketplace_models

from . import PLUGIN_NAME

logger = logging.getLogger(__name__)


class TimePeriod(object):
    def __init__(self, start, end):
        if not isinstance(start, datetime.datetime):
            start = parse_datetime(start)

        if not isinstance(end, datetime.datetime):
            end = parse_datetime(end)

        self.start = start
        self.end = end


def is_interval_in_schedules(interval, schedules):
    for s in schedules:
        if interval.start >= s.start:
            if interval.end <= s.end:
                return True

    return False


def get_info_about_upcoming_bookings():
    tomorrow = datetime.date.today() + datetime.timedelta(days=1)
    upcoming_bookings = marketplace_models.Resource.objects.filter(
        offering__type=PLUGIN_NAME,
        state=marketplace_models.Resource.States.OK,
        attributes__schedules__0__start__icontains='%s-%02d-%02dT' % (tomorrow.year, tomorrow.month, tomorrow.day))

    result = []

    for resource in upcoming_bookings:
        try:
            order_item = marketplace_models.OrderItem.objects.get(
                resource=resource,
                type=marketplace_models.OrderItem.Types.CREATE)
            user = order_item.order.created_by
        except marketplace_models.OrderItem.DoesNotExist:
            logger.warning('Skipping notification because '
                           'marketplace resource hasn\'t got a order item. '
                           'Resource ID: %s', resource.id)
        except marketplace_models.OrderItem.MultipleObjectsReturned:
            logger.warning('Skipping notification because '
                           'marketplace resource has got few order items. '
                           'Resource ID: %s', resource.id)
        else:
            if filter(lambda x: x['user'] == user, result):
                filter(lambda x: x['user'] == resource.project.customer, result)[0]['resources'].append(resource)
            else:
                result.append({
                    'user': user,
                    'resources': [resource],
                })

    return result



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-07-14 00:19
from __future__ import unicode_literals

from django.db import migrations, models


def migrate_offering_price(apps, schema_editor):
    Offering = apps.get_model('support', 'Offering')
    for offering in Offering.objects.all():
        offering.unit_price = offering.price
        offering.save(update_fields=['unit_price'])


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0010_add_unit_price'),
    ]

    operations = [
        migrations.RunPython(migrate_offering_price, elidable=True),
        migrations.RemoveField(
            model_name='offering',
            name='price',
        ),
    ]



class JupyterHubManagementConstants(object):
    SYNC_CONFIGURATION = 'install_jupyter'
    MAKE_VIRTUAL_ENV_GLOBAL = 'make_virtual_env_global'
    MAKE_VIRTUAL_ENV_LOCAL = 'make_virtual_env_local'
    DELETE_JUPYTER_HUB = 'delete_jupyter_hub'



from rest_framework.exceptions import APIException
from rest_framework.status import HTTP_423_LOCKED
from waldur_ansible.jupyter_hub_management.backend import locking_service
from waldur_ansible.python_management import models as python_management_models, utils as python_management_utils

from waldur_core.core import models as core_models
from . import models, executors


class JupyterHubManagementService(object):
    executor = executors.JupyterHubManagementRequestExecutor

    def schedule_jupyter_hub_management_removal(self, persisted_jupyter_hub_management):
        delete_request = models.JupyterHubManagementDeleteRequest(jupyter_hub_management=persisted_jupyter_hub_management)

        if not locking_service.JupyterHubManagementBackendLockingService.is_processing_allowed(delete_request):
            raise APIException(code=HTTP_423_LOCKED)

        delete_request.save()

        self.executor.execute(delete_request, async=True)

    def issue_localize_globalize_requests(self, updated_jupyter_hub_management, validated_data):
        virtual_environments = validated_data['updated_virtual_environments']
        virtual_environments_to_localize = []
        virtual_environments_to_globalize = []
        for virtual_environment in virtual_environments:
            persisted_virtual_environment = python_management_models.VirtualEnvironment.objects.get(
                python_management=updated_jupyter_hub_management.python_management, name=virtual_environment['name'])
            if persisted_virtual_environment.jupyter_hub_global is not virtual_environment['jupyter_hub_global']:
                if virtual_environment['jupyter_hub_global']:
                    virtual_environments_to_globalize.append(virtual_environment)
                else:
                    virtual_environments_to_localize.append(virtual_environment)

        for virtual_environment_to_globalize in virtual_environments_to_globalize:
            globalize_request = models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest(
                jupyter_hub_management=updated_jupyter_hub_management, virtual_env_name=virtual_environment_to_globalize['name'])
            self.execute_or_refuse_request(globalize_request)

        for virtual_environment_to_localize in virtual_environments_to_localize:
            localize_request = models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest(
                jupyter_hub_management=updated_jupyter_hub_management, virtual_env_name=virtual_environment_to_localize['name'])
            self.execute_or_refuse_request(localize_request)

    def execute_or_refuse_request(self, localize_request):
        if locking_service.JupyterHubManagementBackendLockingService.is_processing_allowed(localize_request):
            localize_request.save()
            self.executor.execute(localize_request, async=True)

    def execute_sync_configuration_request_if_allowed(self, persisted_jupyter_hub_management):
        sync_config_request = models.JupyterHubManagementSyncConfigurationRequest(jupyter_hub_management=persisted_jupyter_hub_management)
        if not locking_service.JupyterHubManagementBackendLockingService.is_processing_allowed(sync_config_request):
            raise APIException(code=HTTP_423_LOCKED)
        sync_config_request.save()
        self.executor.execute(sync_config_request, async=True)

    def has_jupyter_hub_config_changed(self, incoming_validated_data, persisted_jupyter_hub_management):
        removed_jupyter_hub_users = self.find_removed_users(persisted_jupyter_hub_management.jupyter_hub_users.all(), incoming_validated_data.get('jupyter_hub_users'))
        if removed_jupyter_hub_users:
            return True
        for jupyter_hub_user in incoming_validated_data.get('jupyter_hub_users'):
            persisted_jupyter_hub_user = self.find_corresponding_persisted_jupyter_hub_user(jupyter_hub_user['username'], persisted_jupyter_hub_management)
            if persisted_jupyter_hub_user is None \
                    or persisted_jupyter_hub_user.admin != jupyter_hub_user['admin'] \
                    or jupyter_hub_user['password'] \
                    or persisted_jupyter_hub_user.whitelisted != jupyter_hub_user['whitelisted']:
                return True
        root_model_changed = incoming_validated_data.get('session_time_to_live_hours') != persisted_jupyter_hub_management.session_time_to_live_hours
        if not root_model_changed:
            persisted_oauth_config = persisted_jupyter_hub_management.jupyter_hub_oauth_config
            if persisted_oauth_config:
                incoming_oauth_config = incoming_validated_data.get('jupyter_hub_oauth_config')
                return persisted_oauth_config.type != incoming_oauth_config.get('type') \
                    or persisted_oauth_config.oauth_callback_url != incoming_oauth_config.get('oauth_callback_url') \
                    or persisted_oauth_config.client_id != incoming_oauth_config.get('client_id') \
                    or persisted_oauth_config.client_secret != incoming_oauth_config.get('client_secret') \
                    or persisted_oauth_config.tenant_id != incoming_oauth_config.get('tenant_id') \
                    or persisted_oauth_config.gitlab_host != incoming_oauth_config.get('gitlab_host')
        else:
            return True

    def find_removed_users(self, persisted_jupyter_hub_users, jupyter_hub_users):
        result = []
        for persisted_jupyter_hub_user in persisted_jupyter_hub_users:
            if not filter(lambda u: u['username'] == persisted_jupyter_hub_user.username, jupyter_hub_users):
                result.append(persisted_jupyter_hub_user)
        return result

    def find_corresponding_persisted_jupyter_hub_user(self, username, jupyter_hub_management):
        return python_management_utils.execute_safely(lambda: models.JupyterHubUser.objects.get(username=username, jupyter_hub_management=jupyter_hub_management))

    def is_last_sync_request_erred(self, persisted_jupyter_hub_management):
        last_sync_config_request = python_management_utils.execute_safely(
            lambda: models.JupyterHubManagementSyncConfigurationRequest.objects.filter(jupyter_hub_management=persisted_jupyter_hub_management).latest('id'))
        return last_sync_config_request.state == core_models.StateMixin.States.ERRED if last_sync_config_request else True



import datetime
import dateutil.parser
import decimal
import paypalrestsdk as paypal
import urlparse
import urllib
import urllib2

from django.conf import settings
from django.core.files.base import ContentFile
from django.utils import six, timezone


class PayPalError(Exception):
    pass


class PaypalPayment(object):
    def __init__(self, payment_id, approval_url, token):
        self.payment_id = payment_id
        self.approval_url = approval_url
        self.token = token


class PaypalBackend(object):

    BACKEND_SERVERS_MAP = {
        'sandbox': 'https://www.sandbox.paypal.com',
        'live': 'https://www.paypal.com',
    }

    def __init__(self):
        config = settings.WALDUR_PAYPAL['BACKEND']
        self.configure(**config)

    def configure(self, mode, client_id, client_secret, currency_name, **kwargs):
        # extra method to validate required config options
        self.currency_name = currency_name

        paypal.configure({
            'mode': mode,
            'client_id': client_id,
            'client_secret': client_secret
        })

        self.server = self.BACKEND_SERVERS_MAP[mode]

    def get_payment_view_url(self, backend_invoice_id, params=None):
        invoice_url = '%s/invoice/payerView/details/%s' % (self.server, backend_invoice_id)
        if params:
            query_params = urllib.urlencode(params)
            invoice_url = '%s?%s' % (invoice_url, query_params)

        return invoice_url

    def _find_approval_url(self, links):
        for link in links:
            if link.rel == 'approval_url':
                return link.href
        raise PayPalError('Approval URL is not found')

    def _find_token(self, approval_url):
        parts = urlparse.urlparse(approval_url)
        params = urlparse.parse_qs(parts.query)
        token = params.get('token')
        if not token:
            raise PayPalError('Unable to parse token from approval_url')
        return token[0]

    def create_invoice(self, invoice):
        """
        Creates invoice with invoice items in it.
        https://developer.paypal.com/docs/api/invoicing/#definition-payment_term
        :param invoice: instance of Invoice class.
        :return: instance of Invoice with backend_id filled.
        """

        if invoice.backend_id:
            return

        if not invoice.items.count():
            raise PayPalError('"items" size must be between 1 and 100.')

        if not invoice.price:
            raise PayPalError('The total cost must not be zero.')

        phone = invoice.issuer_details.get('phone', {})
        if not phone:
            raise PayPalError('"phone" is a required attribute')

        if phone and 'country_code' not in phone:
            raise PayPalError('"phone"."country_code" is a required attribute')

        if phone and 'national_number' not in phone:
            raise PayPalError('"phone"."national_number" is a required attribute')

        invoice_details = {
            'merchant_info': {
                'email': invoice.issuer_details.get('email'),
                'business_name': invoice.issuer_details.get('company'),
                'phone': {
                    'country_code': phone.get('country_code'),
                    'national_number': phone.get('national_number'),
                },
                'address': {
                    'line1': invoice.issuer_details.get('address'),
                    'city': invoice.issuer_details.get('city'),
                    'state': invoice.issuer_details.get('state'),
                    'postal_code': invoice.issuer_details.get('postal'),
                    'country_code': invoice.issuer_details.get('country_code')
                }
            },
            'items': [
                {
                    'name': item.name,
                    'unit_of_measure': item.unit_of_measure,
                    'quantity': item.quantity,
                    'date': self._format_date(item.start.date()),
                    'unit_price': {
                        'currency': self.currency_name,
                        'value': self._format_decimal(item.unit_price),
                    }
                } for item in invoice.items.iterator()
            ],
            'tax_inclusive': False,
            'payment_term': {
                'due_date': self._format_date(invoice.end_date),
            },
            'total_amount': {
                'currency': self.currency_name,
                'value': self._format_decimal(invoice.total)
            }
            # 'logo_url': pass logo url if needed. 250x90, HTTPS. Image is not displayed o PDF atm.
        }

        if invoice.tax_percent and invoice.tax_percent > 0:
            for item in invoice_details['items']:
                item['tax'] = {
                    'name': 'VAT',
                    'percent': self._format_decimal(invoice.tax_percent),
                }

        invoice_details['billing_info'] = [
            {
                'email': invoice.customer.email,
                'business_name': invoice.customer.name,
            }
        ]

        backend_invoice = paypal.Invoice(invoice_details)

        try:
            if backend_invoice.create():
                invoice.state = backend_invoice.status
                invoice.backend_id = backend_invoice.id
                invoice.number = backend_invoice.number
                invoice.save(update_fields=['state', 'backend_id', 'number'])

                return invoice
            else:
                raise PayPalError(backend_invoice.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def send_invoice(self, invoice):
        if invoice.state != invoice.States.DRAFT:
            raise PayPalError('Invoice must be in "%s" state' % invoice.States.DRAFT)

        try:
            backend_invoice = paypal.Invoice.find(invoice.backend_id)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

        if not backend_invoice.send():
            raise PayPalError(backend_invoice.error)

        return invoice

    def pull_invoice(self, invoice):
        try:
            backend_invoice = paypal.Invoice.find(invoice.backend_id)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

        invoice.state = backend_invoice.status
        invoice.number = backend_invoice.number
        invoice.save(update_fields=['state', 'number'])
        return invoice

    def make_payment(self, amount, tax, description, return_url, cancel_url):
        """
        Make PayPal payment using Express Checkout workflow.
        https://developer.paypal.com/docs/api/payments/

        :param amount: Decimal value of payment including VAT tax.
        :param tax: Decimal value of VAT tax.
        :param description: Description of payment.
        :param return_url: Callback view URL for approved payment.
        :param cancel_url: Callback view URL for cancelled payment.
        :return: Object containing backend payment id, approval URL and token.
        """
        if amount < tax:
            raise PayPalError('Payment amount should be greater than tax.')

        payment = paypal.Payment({
            'intent': 'sale',
            'payer': {'payment_method': 'paypal'},
            'transactions': [
                {
                    'amount': {
                        'total': self._format_decimal(amount),
                        'currency': self.currency_name,
                        'details': {
                            'subtotal': self._format_decimal(amount - tax),
                            'tax': self._format_decimal(tax)
                        }
                    },
                    'description': description
                }
            ],
            'redirect_urls': {
                'return_url': return_url,
                'cancel_url': cancel_url
            }
        })

        try:
            if payment.create():
                approval_url = self._find_approval_url(payment.links)
                token = self._find_token(approval_url)
                return PaypalPayment(payment.id, approval_url, token)
            else:
                raise PayPalError(payment.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def approve_payment(self, payment_id, payer_id):
        try:
            payment = paypal.Payment.find(payment_id)
            # When payment is not found PayPal returns empty result instead of raising an exception
            if not payment:
                raise PayPalError('Payment not found')
            if payment.execute({'payer_id': payer_id}):
                return True
            else:
                raise PayPalError(payment.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def create_plan(self, amount, tax, name, description, return_url, cancel_url):
        """
        Create and activate monthly billing plan.
        https://developer.paypal.com/docs/api/payments.billing-plans

        :param amount: Decimal value of plan payment for one month including tax.
        :param tax: Decimal value of VAT tax.
        :param name: Name of the billing plan.
        :param description: Description of the billing plan.
        :param return_url: Callback view URL for approved billing plan.
        :param cancel_url: Callback view URL for cancelled billing plan.
        :return: Billing plan ID.
        """
        if amount < tax:
            raise PayPalError('Plan price should be greater than tax.')

        plan = paypal.BillingPlan({
            'name': name,
            'description': description,
            'type': 'INFINITE',
            'payment_definitions': [{
                'name': 'Monthly payment for {}'.format(name),
                'type': 'REGULAR',
                'frequency_interval': 1,
                'frequency': 'MONTH',
                'cycles': 0,
                'amount': {
                    'currency': self.currency_name,
                    'value': self._format_decimal(amount - tax)
                },
                'charge_models': [
                    {
                        'type': 'TAX',
                        'amount': {
                            'currency': self.currency_name,
                            'value': self._format_decimal(tax)
                        }
                    }
                ]
            }],
            'merchant_preferences': {
                'return_url': return_url,
                'cancel_url': cancel_url,
                'auto_bill_amount': 'YES',
            }
        })

        try:
            if plan.create() and plan.activate():
                return plan.id
            else:
                raise PayPalError(plan.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def _format_decimal(self, value):
        """
        PayPal API expects at most two decimal places with a period separator.
        """
        return "%.2f" % value

    def _format_date(self, date):
        """
        At the moment timezone is ignored as only days of resources usage are counted, not hours.
        """
        return date.strftime('%Y-%m-%d UTC')

    def create_agreement(self, plan_id, name):
        """
        Create billing agreement. On success returns approval_url and token
        """
        # PayPal does not support immediate start of agreement
        # That's why we need to increase start date by small amount of time
        start_date = datetime.datetime.utcnow() + datetime.timedelta(minutes=1)

        # PayPal does not fully support ISO 8601 format
        formatted_date = start_date.strftime('%Y-%m-%dT%H:%M:%SZ')

        agreement = paypal.BillingAgreement({
            'name': name,
            'description': 'Agreement for {}'.format(name),
            'start_date': formatted_date,
            'payer': {'payment_method': 'paypal'},
            'plan': {'id': plan_id}
        })
        try:
            if agreement.create():
                approval_url = self._find_approval_url(agreement.links)

                # PayPal does not return agreement ID until it is approved
                # That's why we need to extract token in order to identify it with agreement in DB
                token = self._find_token(approval_url)
                return approval_url, token
            else:
                raise PayPalError(agreement.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def execute_agreement(self, payment_token):
        """
        Agreement should be executed if user has approved it.
        On success returns agreement id
        """
        try:
            agreement = paypal.BillingAgreement.execute(payment_token)
            if not agreement:
                raise PayPalError('Can not execute agreement')
            return agreement.id
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def get_agreement(self, agreement_id):
        """
        Get agreement from PayPal by ID
        """
        try:
            agreement = paypal.BillingAgreement.find(agreement_id)
            # When agreement is not found PayPal returns empty result instead of raising an exception
            if not agreement:
                raise PayPalError('Agreement not found')
            return agreement
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def cancel_agreement(self, agreement_id):
        agreement = self.get_agreement(agreement_id)

        try:
            # Because user may cancel agreement via PayPal web UI
            # we need to distinguish it from cancel done via API
            if agreement.cancel({'note': 'Canceling the agreement by application'}):
                return True
            else:
                raise PayPalError(agreement.error)
        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def get_agreement_transactions(self, agreement_id, start_date, end_date=None):
        if not end_date:
            end_date = timezone.now()

        # If start and end date are the same PayPal raises exceptions
        # That's why we need to increase end_date by one day
        if end_date - start_date < datetime.timedelta(days=1):
            end_date += datetime.timedelta(days=1)

        formatted_start_date = start_date.strftime('%Y-%m-%d')
        formatted_end_date = end_date.strftime('%Y-%m-%d')

        agreement = self.get_agreement(agreement_id)
        try:
            data = agreement.search_transactions(formatted_start_date, formatted_end_date)
            txs = data.agreement_transaction_list
            if not txs:
                return []

            results = []
            for tx in txs:
                if tx.status != 'Completed':
                    continue
                results.append({
                    'time_stamp': dateutil.parser.parse(tx.time_stamp),
                    'transaction_id': tx.transaction_id,
                    'amount': decimal.Decimal(tx.amount.value),
                    'payer_email': tx.payer_email
                })
            return results

        except paypal.exceptions.ConnectionError as e:
            six.reraise(PayPalError, e)

    def download_invoice_pdf(self, invoice):
        if not invoice.backend_id:
            raise PayPalError('Invoice for date %s and customer %s could not be found' % (
                invoice.invoice_date.strftime('%Y-%m-%d'),
                invoice.customer.name,
            ))

        invoice_url = self.get_payment_view_url(invoice.backend_id, {'printPdfMode': 'true'})
        response = urllib2.urlopen(invoice_url)  # nosec
        content = response.read()
        invoice.pdf.save(invoice.file_name, ContentFile(content), save=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:09
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0033_unique_instance_backend_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instance',
            name='action_details',
            field=waldur_core.core.fields.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='snapshot',
            name='action_details',
            field=waldur_core.core.fields.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='subnet',
            name='allocation_pools',
            field=waldur_core.core.fields.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='subnet',
            name='dns_nameservers',
            field=waldur_core.core.fields.JSONField(default=list, help_text='List of DNS name servers associated with the subnet.'),
        ),
        migrations.AlterField(
            model_name='volume',
            name='action_details',
            field=waldur_core.core.fields.JSONField(default=dict),
        ),
    ]



TIME_SUFFIXES = {
    's': 1,
    'm': 60,
    'h': 3600,
    'd': 86400,
    'w': 604800,
}


def parse_time(value):
    """
    Convert 1d to 86400 seconds.
    """
    if value.isdigit():
        return int(value)

    if not value[:-1].isdigit():
        raise ValueError('Invalid time value %s' % value)

    for suffix, factor in TIME_SUFFIXES.items():
        if not value.endswith(suffix):
            continue

        stripped = value[:-1]
        if stripped.isdigit():
            return int(stripped) * factor

    raise ValueError('Invalid time value %s' % value)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-03-02 16:08
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0022_floatingip_internal_ip'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='instance',
            name='external_ip',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-08-21 12:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0021_remove_default_draft_state_from_order'),
    ]

    operations = [
        migrations.AddField(
            model_name='section',
            name='is_standalone',
            field=models.BooleanField(default=False, help_text='Whether section is rendered as a separate tab.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-20 16:12
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0017_genericinvoiceitem'),
    ]

    operations = [
        migrations.AddField(
            model_name='genericinvoiceitem',
            name='quantity',
            field=models.PositiveIntegerField(default=0),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-13 08:28
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_mastermind.invoices.utils


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('structure', '0052_customer_subnets'),
        ('invoices', '0016_remove_daily_price'),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericInvoiceItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[(b'month', 'Per month'), (b'half_month', 'Per half month'), (b'day', 'Per day'), (b'hour', 'Per hour'), (b'quantity', 'Quantity')], default=b'day', max_length=30)),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_start, help_text='Date and time when item usage has started.')),
                ('end', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_end, help_text='Date and time when item usage has ended.')),
                ('project_name', models.CharField(blank=True, max_length=150)),
                ('project_uuid', models.CharField(blank=True, max_length=32)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('details', waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Stores data about scope')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='generic_items', to='invoices.Invoice')),
                ('project', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='structure.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from celery import shared_task

from django.contrib.contenttypes import models as ct_models
from django.utils import timezone

from waldur_core.quotas import models as quota_models
from waldur_core.structure import models as structure_models

from . import cost_tracking, openstack, slurm, utils, models


@shared_task(name='analytics.push_points')
def push_points():
    client = utils.get_influxdb_client()
    if not client:
        return
    points = []
    points.extend(openstack.get_tenants())
    points.extend(openstack.get_instances())
    points.extend(cost_tracking.get_total_cost())
    points.extend(slurm.get_usage())
    utils.write_points(client, points)


@shared_task(name='analytics.sync_daily_quotas')
def sync_daily_quotas():
    date = timezone.now().date()
    for model in (structure_models.Project, structure_models.Customer):
        content_type = ct_models.ContentType.objects.get_for_model(model)
        for quota in quota_models.Quota.objects.filter(content_type=content_type):
            if not quota.scope:
                continue
            models.DailyQuotaHistory.objects.update_or_create_quota(
                quota.scope, quota.name, date, quota.usage)



import logging

from celery import shared_task
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from python_freeipa import exceptions as freeipa_exceptions

from . import models, utils
from .backend import FreeIPABackend


logger = logging.getLogger(__name__)


def schedule_sync():
    """
    This function calls task only if it is not already running.
    The goal is to avoid race conditions during concurrent task execution.
    """
    if utils.is_syncing():
        logger.debug('Skipping FreeIPA synchronization because synchronization is already in progress.')
        return

    if not settings.WALDUR_FREEIPA['ENABLED']:
        logger.debug('Skipping FreeIPA synchronization because plugin is disabled.')
        return

    if not settings.WALDUR_FREEIPA['GROUP_SYNCHRONIZATION_ENABLED']:
        logger.debug('Skipping FreeIPA group synchronization because this feature is disabled.')
        return

    utils.renew_task_status()
    _sync_groups.apply_async(countdown=10)


@shared_task(name='waldur_freeipa.sync_groups')
def sync_groups():
    """
    This task is used by Celery beat in order to periodically
    schedule FreeIPA group synchronization.
    """
    schedule_sync()


@shared_task()
def _sync_groups():
    """
    This task actually calls backend. It is called asynchronously
    either by signal handler or Celery beat schedule.
    """
    FreeIPABackend().synchronize_groups()


def schedule_sync_names():
    _sync_names.apply_async(countdown=10)


@shared_task()
def _sync_names():
    FreeIPABackend().synchronize_names()


def schedule_sync_gecos():
    _sync_gecos.apply_async(countdown=10)


@shared_task()
def _sync_gecos():
    FreeIPABackend().synchronize_gecos()


@shared_task(name='waldur_freeipa.sync_profile_ssh_keys')
def sync_profile_ssh_keys(profile_id):
    try:
        profile = models.Profile.objects.get(id=profile_id)
    except ObjectDoesNotExist:
        logger.debug('Skipping SSH key synchronization because FreeIPA profile has been deleted. '
                     'Profile ID: %s', profile_id)
        return

    try:
        FreeIPABackend().update_ssh_keys(profile)
    except freeipa_exceptions.NotFound:
        logger.warning('Skipping SSH key synchronization because '
                       'FreeIPA profile has been removed on backend. '
                       'Profile ID: %s', profile.id)
        return



from __future__ import unicode_literals

import logging

from waldur_core.cost_tracking import CostTrackingStrategy, ConsumableItem, CostTrackingRegister
from waldur_openstack.openstack import models as openstack_models

from . import models, utils


logger = logging.getLogger(__name__)


class TenantStrategy(CostTrackingStrategy):
    resource_class = openstack_models.Tenant

    @classmethod
    def get_consumable_items(cls):
        for package_template in models.PackageTemplate.objects.all():
            yield utils.get_consumable_item(package_template)

    @classmethod
    def get_configuration(cls, tenant):
        configuration = {}
        if tenant.state != tenant.States.ERRED:
            if 'package_name' not in tenant.extra_configuration:
                logger.debug(
                    'Package name is not defined in configuration of tenant %s, (PK: %s)', tenant.name, tenant.pk)
            else:
                package_name = tenant.extra_configuration['package_name']
                configuration = {
                    ConsumableItem(item_type=utils.Types.PACKAGE_TEMPLATE, key=package_name): 1,
                }
        return configuration


CostTrackingRegister.register_strategy(TenantStrategy)



"""
See also:
https://docs.microsoft.com/en-us/azure/architecture/best-practices/naming-conventions
https://docs.microsoft.com/en-us/azure/virtual-machines/linux/faq
"""

import re

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator, MinLengthValidator, \
    MaxLengthValidator, MinValueValidator, MaxValueValidator
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.validators import BlacklistValidator


ResourceGroupNameValidator = RegexValidator(
    regex=re.compile(r'^[-\w._()]+$'),
    message=_('The name can include alphanumeric, underscore, parentheses, '
              'hyphen, period (except at end), and Unicode characters '
              'that match the allowed characters.')
)


VirtualMachineNameValidator = RegexValidator(
    regex=re.compile(r'^[a-zA-Z][a-zA-Z0-9-]{0,13}[a-zA-Z0-9]$'),
    message=_('The name can contain only letters, numbers, and hyphens. '
              'The name must be shorter than 15 characters and start with '
              'a letter and must end with a letter or a number.')
)


class VirtualMachineUsernameValidator(BlacklistValidator):
    blacklist = (
        'administrator',
        'admin',
        'user',
        'user1',
        'test',
        'user2',
        'test1',
        'user3',
        'admin1',
        '1',
        '123',
        'a',
        'actuser',
        'adm',
        'admin2',
        'aspnet',
        'backup',
        'console',
        'david',
        'guest',
        'john',
        'owner',
        'root',
        'server',
        'sql',
        'support',
        'support_388945a0',
        'sys',
        'test2',
        'test3',
        'user4',
        'user5',
    )


class VirtualMachinePasswordValidator(BlacklistValidator):
    blacklist = (
        'abc@123',
        'P@$$w0rd',
        'P@ssw0rd',
        'P@ssword123',
        'Pa$$word',
        'pass@word1',
        'Password!',
        'Password1',
        'Password22',
        'iloveyou!',
    )


def validate_password(password):
    groups = (r'[a-z]', r'[A-Z]', r'[0-9]', r'[\W_]')
    if sum(bool(re.search(g, password)) for g in groups) < 3:
        raise ValidationError(_('The supplied password must contain 3 of the following: '
                                'a lowercase character, an uppercase character, a number, '
                                'a special character.'))


VirtualMachinePasswordValidators = [
    MinLengthValidator(6),
    MaxLengthValidator(72),
    validate_password,
    VirtualMachinePasswordValidator,
]


NetworkingNameValidator = RegexValidator(
    regex=re.compile(r'^[a-zA-Z][a-zA-Z0-9._-]+$'),
    message=_('The name can contain only letters, numbers, underscore, period and hyphens.')
)


StorageAccountNameValidator = RegexValidator(
    regex=re.compile(r'^[a-z][a-z0-9]{2,23}$'),
    message=_('The name can contain only letters and numbers.')
)


SQLServerNameValidator = RegexValidator(
    regex=re.compile(r'^[a-z0-9][a-z0-9-]+[a-z0-9]$'),
    message=_('The name can only be made up of lowercase letters "a"-"z", the numbers 0-9 and the hyphen. '
              'The hyphen may not lead or trail in the name.')
)


class SQLServerUsernameValidator(BlacklistValidator):
    blacklist = (
        'azure_superuser',
        'admin',
        'administrator',
        'root',
        'guest',
        'public',
    )


SQLServerPasswordValidators = [
    MinLengthValidator(8),
    MaxLengthValidator(128),
    validate_password,
]


# See also: https://docs.microsoft.com/en-us/azure/postgresql/concepts-pricing-tiers
SQLServerStorageValidators = [
    MinValueValidator(5 * 1024),
    MaxValueValidator(4 * 1024 * 1024),
]



from argparse import FileType
import operator
import json
from sys import stdin

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand

from waldur_auth_openid.log import OpenIDEventLogger
from waldur_auth_saml2.log import Saml2AuthEventLogger
from waldur_auth_social.log import SocialEventLogger
from waldur_core.core.log import AuthEventLogger
from waldur_core.core.models import User
from waldur_core.logging.models import Event, Feed
from waldur_core.structure import SupportedServices
from waldur_core.structure.models import Customer, Project
from waldur_freeipa.log import FreeIPAEventLogger
from waldur_mastermind.common.utils import parse_datetime
from waldur_mastermind.marketplace.models import Resource

SKIPPED_FIELDS = (
    'importance',
    'importance_code',
    'logger',
    'type',
    '@version',
    'host',
    'levelname',
    'tags',
    'port',
)

USER_LOGGERS = (
    OpenIDEventLogger,
    Saml2AuthEventLogger,
    SocialEventLogger,
    AuthEventLogger,
    FreeIPAEventLogger,
)

USER_EVENTS = reduce(operator.add, map(lambda logger: tuple(logger.Meta.event_types), USER_LOGGERS))

RESOURCE_MODELS = SupportedServices.get_resource_models()


def get_scopes(event):
    scopes = []
    user_uuid = event.context.get('user_uuid')
    if user_uuid and event.event_type in USER_EVENTS:
        try:
            user = User.objects.get(uuid=user_uuid)
            scopes.append(user)
        except ObjectDoesNotExist:
            pass

    affected_user_uuid = event.context.get('affected_user_uuid')
    if affected_user_uuid:
        try:
            user = User.objects.get(uuid=affected_user_uuid)
            scopes.append(user)
        except ObjectDoesNotExist:
            pass

    customer_uuid = event.context.get('customer_uuid')
    if customer_uuid:
        try:
            customer = Customer.objects.get(uuid=customer_uuid)
            scopes.append(customer)
        except ObjectDoesNotExist:
            pass

    project_uuid = event.context.get('project_uuid')
    if project_uuid:
        try:
            project = Project.objects.get(uuid=project_uuid)
            scopes.append(project)
        except ObjectDoesNotExist:
            pass

    resource_uuid = event.context.get('resource_uuid')
    if resource_uuid:
        if event.event_type.startswith('marketplace'):
            model = Resource
        else:
            resource_type = event.context.get('resource_type')
            model = RESOURCE_MODELS.get(resource_type)

        if model:
            try:
                resource = model.objects.get(uuid=resource_uuid)
                scopes.append(resource)
            except ObjectDoesNotExist:
                pass

    return scopes


class Command(BaseCommand):
    help = 'Import ElasticSearch events to PostgreSQL database.'

    def add_arguments(self, parser):
        parser.add_argument('input', nargs='?', type=FileType('r'),
                            default=stdin)

    def handle(self, *args, **options):
        counter = 0
        for line in options['input']:
            record = json.loads(line)
            record = record['_source']
            for field in SKIPPED_FIELDS:
                record.pop(field, None)
            created = parse_datetime(record.pop('@timestamp'))
            event_type = record.pop('event_type')
            message = record.pop('message')
            event = Event.objects.create(
                created=created,
                event_type=event_type,
                message=message,
                context=record,
            )
            for scope in get_scopes(event):
                Feed.objects.create(scope=scope, event=event)
            counter += 1
            if counter % 100 == 0:
                self.stdout.write('%s events have been imported' % counter)
        self.stdout.write('%s events have been imported' % counter)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-18 09:28
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_vmware', '0013_add_folder'),
    ]

    operations = [
        migrations.AddField(
            model_name='virtualmachine',
            name='folder',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='waldur_vmware.Folder'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-07 07:31
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0018_serviceprovider_description'),
    ]

    operations = [
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[('month', 'Per month'), ('half_month', 'Per half month'), ('day', 'Per day'), ('hour', 'Per hour'), ('quantity', 'Quantity')], default='day', max_length=30)),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plans', to='marketplace.Offering')),
            ],
            options={
                'abstract': False,
                'ordering': ('name',)
            },
        ),
        migrations.AddField(
            model_name='orderitem',
            name='plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from waldur_core.structure.permissions import _get_project


def set_invoice_item_project(apps, schema_editor):
    OpenStackItem = apps.get_model('invoices', 'OpenStackItem')
    OfferingItem = apps.get_model('invoices', 'OfferingItem')

    for item in OpenStackItem.objects.all().exclude(package__isnull=True):
        project = _get_project(item.package)
        item.project_name = project.name
        item.project_uuid = project.uuid.hex
        item.save(update_fields=['project_name', 'project_uuid'])

    for item in OfferingItem.objects.all().exclude(offering__isnull=True):
        item.project_name = item.offering.project.name
        item.project_uuid = item.offering.project.uuid.hex
        item.save(update_fields=['project_name', 'project_uuid'])


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0012_add_project_to_invoice_item'),
    ]

    operations = [
        migrations.RunPython(set_invoice_item_project, elidable=True),
    ]



from __future__ import unicode_literals, division

import StringIO
import base64
from datetime import timedelta
import decimal
import logging
from calendar import monthrange

import datetime
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.core import models as core_models
from waldur_core.core import utils as core_utils
from waldur_core.core.exceptions import IncorrectStateException
from waldur_core.structure import models as structure_models
from waldur_mastermind.common import mixins as common_mixins
from waldur_mastermind.common.utils import quantize_price
from waldur_mastermind.invoices.utils import get_price_per_day
from waldur_mastermind.packages import models as package_models

from . import managers, utils

logger = logging.getLogger(__name__)

Units = common_mixins.UnitPriceMixin.Units


@python_2_unicode_compatible
class Invoice(core_models.UuidMixin, models.Model):
    """ Invoice describes billing information about purchased resources for customers on a monthly basis """

    class Permissions(object):
        customer_path = 'customer'

    class Meta(object):
        unique_together = ('customer', 'month', 'year')

    class States(object):
        PENDING = 'pending'
        CREATED = 'created'
        PAID = 'paid'
        CANCELED = 'canceled'

        CHOICES = ((PENDING, _('Pending')), (CREATED, _('Created')), (PAID, _('Paid')), (CANCELED, _('Canceled')))

    month = models.PositiveSmallIntegerField(default=utils.get_current_month,
                                             validators=[MinValueValidator(1), MaxValueValidator(12)])
    year = models.PositiveSmallIntegerField(default=utils.get_current_year)
    state = models.CharField(max_length=30, choices=States.CHOICES, default=States.PENDING)
    customer = models.ForeignKey(structure_models.Customer,
                                 verbose_name=_('organization'),
                                 related_name='+',
                                 on_delete=models.CASCADE)
    current_cost = models.DecimalField(default=0, max_digits=10, decimal_places=2,
                                       help_text=_('Cached value for current cost.'),
                                       editable=False)
    tax_percent = models.DecimalField(default=0, max_digits=4, decimal_places=2,
                                      validators=[MinValueValidator(0), MaxValueValidator(100)])
    invoice_date = models.DateField(null=True, blank=True,
                                    help_text=_('Date then invoice moved from state pending to created.'))
    _file = models.TextField(blank=True, editable=False)

    tracker = FieldTracker()

    def update_current_cost(self):
        total_current = self.total_current

        if self.current_cost != total_current:
            self.current_cost = total_current
            self.save(update_fields=['current_cost'])

    @property
    def tax(self):
        return self.price * self.tax_percent / 100

    @property
    def total(self):
        return self.price + self.tax

    @property
    def price(self):
        return sum((item.price for item in self.items))

    @property
    def tax_current(self):
        return self.price_current * self.tax_percent / 100

    @property
    def total_current(self):
        return self.price_current + self.tax_current

    @property
    def price_current(self):
        return sum((item.price_current for item in self.items))

    @property
    def items(self):
        return self.generic_items.all()

    @property
    def due_date(self):
        if self.invoice_date:
            return self.invoice_date + datetime.timedelta(days=settings.WALDUR_INVOICES['PAYMENT_INTERVAL'])

    @property
    def number(self):
        return 100000 + self.id

    def set_created(self):
        """
        Change state from pending to billed
        """
        if self.state != self.States.PENDING:
            raise IncorrectStateException(_('Invoice must be in pending state.'))

        self.state = self.States.CREATED
        self.invoice_date = timezone.now().date()
        self.save(update_fields=['state', 'invoice_date'])

    @property
    def file(self):
        if not self._file:
            return

        content = base64.b64decode(self._file)
        return StringIO.StringIO(content)

    @file.setter
    def file(self, value):
        self._file = value

    def has_file(self):
        return bool(self._file)

    def get_filename(self):
        return 'invoice_{}.pdf'.format(self.uuid)

    def __str__(self):
        return '%s | %s-%s' % (self.customer, self.year, self.month)


@python_2_unicode_compatible
class InvoiceItem(common_mixins.ProductCodeMixin, common_mixins.UnitPriceMixin):
    """
    It is expected that get_scope_type method is defined as class method in scope class
    as it is used in generic invoice item serializer.
    """
    invoice = models.ForeignKey(Invoice, related_name='generic_items')
    content_type = models.ForeignKey(ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    quantity = models.PositiveIntegerField(default=0)

    scope = GenericForeignKey('content_type', 'object_id')
    name = models.TextField(default='')
    details = JSONField(default=dict, blank=True, help_text=_('Stores data about scope'))

    start = models.DateTimeField(default=utils.get_current_month_start,
                                 help_text=_('Date and time when item usage has started.'))
    end = models.DateTimeField(default=utils.get_current_month_end,
                               help_text=_('Date and time when item usage has ended.'))

    # Project name and UUID should be stored separately because project is not available after removal
    project = models.ForeignKey(structure_models.Project, on_delete=models.SET_NULL, null=True)
    project_name = models.CharField(max_length=150, blank=True)
    project_uuid = models.CharField(max_length=32, blank=True)

    objects = managers.InvoiceItemManager()
    tracker = FieldTracker()

    @property
    def tax(self):
        return self.price * self.invoice.tax_percent / 100

    @property
    def tax_current(self):
        return self.price_current * self.invoice.tax_percent / 100

    @property
    def total(self):
        return self.price + self.tax

    def _price(self, current=False):
        return self.unit_price * decimal.Decimal(self.get_factor(current))

    def get_factor(self, current=False):
        month_days = monthrange(self.start.year, self.start.month)[1]

        if self.unit == self.Units.QUANTITY:
            return self.quantity
        elif self.unit == self.Units.PER_HOUR:
            if current:
                return utils.get_full_hours(self.start, min(self.end, timezone.now()))
            else:
                return utils.get_full_hours(self.start, self.end)
        elif self.unit == self.Units.PER_DAY:
            if current:
                return utils.get_full_days(self.start, min(self.end, timezone.now()))
            else:
                return self.usage_days
        elif self.unit == self.Units.PER_HALF_MONTH:
            if (self.start.day == 1 and self.end.day == 15) or (self.start.day == 16 and self.end.day == month_days):
                return 1
            elif (self.start.day == 1 and self.end.day == month_days):
                return 2
            elif (self.start.day == 1 and self.end.day > 15):
                return quantize_price(1 + (self.end.day - 15) / decimal.Decimal(month_days / 2))
            elif (self.start.day < 16 and self.end.day == month_days):
                return quantize_price(1 + (16 - self.start.day) / decimal.Decimal(month_days / 2))
            else:
                return quantize_price((self.end.day - self.start.day + 1) / decimal.Decimal(month_days / 2.0))
        # By default PER_MONTH
        else:
            if self.start.day == 1 and self.end.day == month_days:
                return 1

            use_days = (self.end - self.start).days + 1
            return quantize_price(decimal.Decimal(use_days) / month_days)

    @property
    def price(self):
        return self._price()

    @property
    def price_current(self):
        return self._price(current=True)

    @property
    def usage_days(self):
        """
        Returns the number of days package was used from the time
        it was purchased or from the start of current month
        """
        full_days = utils.get_full_days(self.start, self.end)
        return full_days

    def terminate(self, end=None):
        self.end = end or timezone.now()
        self.save(update_fields=['end'])

    def __str__(self):
        return self.name or '<InvoiceItem %s>' % self.pk

    def create_compensation(self, name, **kwargs):
        FIELDS = (
            'invoice',
            'project',
            'project_name',
            'project_uuid',
            'product_code',
            'article_code',
            'unit',
            'unit_price',
            'start',
            'end',
        )

        params = {field: getattr(self, field) for field in FIELDS}
        params.update(kwargs)
        if params['unit_price'] > 0:
            params['unit_price'] *= -1
        params['details'] = {
            'name': _('Compensation for downtime. Resource name: %s') % name
        }

        return InvoiceItem.objects.create(**params)


def get_default_downtime_start():
    return timezone.now() - settings.WALDUR_INVOICES['DOWNTIME_DURATION_MINIMAL']


class ServiceDowntime(models.Model):
    """
    Currently this model is restricted to OpenStack package only.
    It is expected that implementation would be generalized to support other resources as well.
    """
    start = models.DateTimeField(
        default=get_default_downtime_start,
        help_text=_('Date and time when downtime has started.')
    )
    end = models.DateTimeField(
        default=timezone.now,
        help_text=_('Date and time when downtime has ended.')
    )
    package = models.ForeignKey(package_models.OpenStackPackage)

    def clean(self):
        self._validate_duration()
        self._validate_offset()
        self._validate_intersection()

    def _validate_duration(self):
        duration = self.end - self.start

        duration_min = settings.WALDUR_INVOICES['DOWNTIME_DURATION_MINIMAL']
        if duration_min is not None and duration < duration_min:
            raise ValidationError(
                _('Downtime duration is too small. Minimal duration is %s') % duration_min
            )

        duration_max = settings.WALDUR_INVOICES['DOWNTIME_DURATION_MAXIMAL']
        if duration_max is not None and duration > duration_max:
            raise ValidationError(
                _('Downtime duration is too big. Maximal duration is %s') % duration_max
            )

    def _validate_offset(self):
        if self.start > timezone.now() or self.end > timezone.now():
            raise ValidationError(
                _('Future downtime is not supported yet. '
                  'Please select date in the past instead.')
            )

    def get_intersection_subquery(self):
        left = Q(start__gte=self.start, start__lte=self.end)
        right = Q(end__gte=self.start, end__lte=self.end)
        inside = Q(start__gte=self.start, end__lte=self.end)
        outside = Q(start__lte=self.start, end__gte=self.end)
        return Q(left | right | inside | outside)

    def _validate_intersection(self):
        qs = ServiceDowntime.objects.filter(self.get_intersection_subquery(), package=self.package)
        if qs.exists():
            ids = ', '.join(str(item.id) for item in qs)
            raise ValidationError(
                _('Downtime period intersects with another period with ID: %s.') % ids
            )


class InvoiceItemAdjuster(object):
    def __init__(self, invoice, source, start, unit_price, unit):
        self.invoice = invoice
        self.source = source
        self.start = start
        self.unit_price = unit_price
        self.unit = unit

    @cached_property
    def content_type(self):
        return ContentType.objects.get_for_model(self.source)

    @property
    def invoice_items(self):
        # TODO: Remove temporary workaround for OpenStack package
        if isinstance(self.source, package_models.OpenStackPackage):
            return InvoiceItem.objects.filter(
                invoice=self.invoice,
                content_type=self.content_type,
                details__tenant_name=self.source.tenant.name,
            )
        return InvoiceItem.objects.filter(
            invoice=self.invoice,
            content_type=self.content_type,
            object_id=self.source.pk,
        )

    @cached_property
    def old_item(self):
        qs = self.invoice_items
        if self.unit == Units.PER_DAY:
            qs = qs.filter(end__day=self.start.day)
        elif self.unit == Units.PER_HOUR:
            qs = qs.filter(end__day=self.start.day, end__hour=self.start.hour)
        elif self.unit == Units.PER_MONTH:
            qs = qs.filter(end__month=self.start.month)
        elif self.unit == Units.PER_HALF_MONTH:
            if self.start.day <= 15:
                qs = qs.filter(end__day__lte=15)
            else:
                qs = qs.filter(end__day__gt=15)
        else:
            qs = qs.none()
        return qs.order_by('-unit_price').first()

    @property
    def old_price(self):
        return get_price_per_day(self.old_item.unit_price, self.old_item.unit)

    @property
    def new_price(self):
        return get_price_per_day(self.unit_price, self.unit)

    def shift_forward(self):
        """
        Adjust old invoice item end field to the end of current unit.
        Adjust new invoice item start field to the start of next unit.
        """
        end = self.old_item.end

        if self.old_item.unit != self.unit and self.unit == Units.PER_MONTH:
            end = core_utils.month_end(end)
        elif self.old_item.unit != self.unit and self.unit == Units.PER_HALF_MONTH:
            if end.day > 15:
                end = core_utils.month_end(end)
            else:
                end = end.replace(day=15)
        elif self.unit == Units.PER_HOUR:
            end = end.replace(minute=59, second=59)
        else:
            end = end.replace(hour=23, minute=59, second=59)

        start = end + timedelta(seconds=1)
        return start, end

    def shift_backward(self):
        """
        Adjust old invoice item end field to the end of previous unit
        Adjust new invoice item field to the start of current unit.
        """
        end = self.old_item.end

        if self.old_item.unit != self.unit and self.unit == Units.PER_MONTH:
            start = core_utils.month_start(end)
        elif self.old_item.unit != self.unit and self.unit == Units.PER_HALF_MONTH:
            if end.day < 15:
                start = core_utils.month_start(end)
            else:
                start = end.replace(day=15)
        elif self.unit == Units.PER_HOUR:
            start = end.replace(minute=0, second=0)
        else:
            start = end.replace(hour=0, minute=0, second=0)

        end = start - timedelta(seconds=1)
        return start, end

    def remove_new_items(self, start):
        """
        Cleanup planned invoice items when new item is created.
        """
        qs = self.invoice_items
        if self.unit == Units.PER_DAY:
            qs = qs.filter(start__day=start.day)
        elif self.unit == Units.PER_HOUR:
            qs = qs.filter(start__day=start.day, start__hour=start.hour)
        else:
            qs = qs.none()

        qs.delete()

    def adjust(self):
        start = self.start

        if self.old_item and self.old_item.price > 0:
            if self.old_price >= self.new_price:
                start, end = self.shift_forward()
            else:
                start, end = self.shift_backward()

            self.old_item.end = end
            self.old_item.save(update_fields=['end'])

        self.remove_new_items(start)

        return start


def adjust_invoice_items(invoice, source, start, unit_price, unit):
    """
    When resource configuration is switched, old invoice item
    is terminated and new invoice item is created.
    In order to avoid double counting we should ensure that
    there're no overlapping invoice items for the same scope.

    By default daily prorate is used even if plan is monthly or half-monthly.
    Two notable exceptions are:
    1) Switching from daily plan to monthly.
    2) Switching between hourly plans.
    """
    return InvoiceItemAdjuster(invoice, source, start, unit_price, unit).adjust()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-26 07:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0007_screenshots'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='native_description',
            field=models.CharField(blank=True, default='', max_length=500),
        ),
        migrations.AddField(
            model_name='offering',
            name='native_name',
            field=models.CharField(blank=True, default='', max_length=160),
        ),
        migrations.AddField(
            model_name='offering',
            name='preferred_language',
            field=models.CharField(blank=True, max_length=10),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_fsm
import waldur_core.structure.models
import waldur_core.core.models
import django.core.validators
import taggit.managers
import waldur_core.logging.loggers
import model_utils.fields
import waldur_core.core.fields
import django.db.models.deletion
import django.utils.timezone
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0035_settings_tags_and_scope'),
        ('taggit', '0002_auto_20150616_2121'),
    ]

    operations = [
        migrations.CreateModel(
            name='Backup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('kept_until', models.DateTimeField(help_text='Guaranteed time of backup retention. If null - keep forever.', null=True, blank=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('state', django_fsm.FSMIntegerField(default=1, choices=[(1, 'Ready'), (2, 'Backing up'), (3, 'Restoring'), (4, 'Deleting'), (5, 'Erred'), (6, 'Deleted')])),
                ('metadata', waldur_core.core.fields.JSONField(help_text='Additional information about backup, can be used for backup restoration or deletion', blank=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.CreateModel(
            name='BackupSchedule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('schedule', waldur_core.core.fields.CronScheduleField(max_length=15, validators=[waldur_core.core.validators.validate_cron_schedule])),
                ('next_trigger_at', models.DateTimeField(null=True)),
                ('timezone', models.CharField(default=django.utils.timezone.get_current_timezone_name, max_length=50)),
                ('is_active', models.BooleanField(default=False)),
                ('retention_time', models.PositiveIntegerField(help_text='Retention time in days')),
                ('maximal_number_of_backups', models.PositiveSmallIntegerField()),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.CreateModel(
            name='Flavor',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('cores', models.PositiveSmallIntegerField(help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(help_text='Root disk size in MiB')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FloatingIP',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('address', models.GenericIPAddressField(protocol='IPv4')),
                ('status', models.CharField(max_length=30)),
                ('backend_id', models.CharField(max_length=255)),
                ('backend_network_id', models.CharField(max_length=255, editable=False)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Floating IP',
                'verbose_name_plural': 'Floating IPs',
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Instance',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('latitude', models.FloatField(null=True, blank=True)),
                ('longitude', models.FloatField(null=True, blank=True)),
                ('key_name', models.CharField(max_length=50, blank=True)),
                ('key_fingerprint', models.CharField(max_length=47, blank=True)),
                ('user_data', models.TextField(help_text='Additional data that will be added to instance on provisioning', blank=True)),
                ('cores', models.PositiveSmallIntegerField(default=0, help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(default=0, help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(default=0, help_text='Disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('external_ips', models.GenericIPAddressField(null=True, protocol='IPv4', blank=True)),
                ('internal_ips', models.GenericIPAddressField(null=True, protocol='IPv4', blank=True)),
                ('image_name', models.CharField(max_length=150, blank=True)),
                ('billing_backend_id', models.CharField(help_text='ID of a resource in backend', max_length=255, blank=True)),
                ('last_usage_update_time', models.DateTimeField(null=True, blank=True)),
                ('state', django_fsm.FSMIntegerField(default=1, help_text='WARNING! Should not be changed manually unless you really know what you are doing.', choices=[(1, 'Provisioning Scheduled'), (2, 'Provisioning'), (3, 'Online'), (4, 'Offline'), (5, 'Starting Scheduled'), (6, 'Starting'), (7, 'Stopping Scheduled'), (8, 'Stopping'), (9, 'Erred'), (10, 'Deletion Scheduled'), (11, 'Deleting'), (13, 'Resizing Scheduled'), (14, 'Resizing'), (15, 'Restarting Scheduled'), (16, 'Restarting')])),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('system_volume_id', models.CharField(max_length=255, blank=True)),
                ('system_volume_size', models.PositiveIntegerField(default=0, help_text='Root disk size in MiB')),
                ('data_volume_id', models.CharField(max_length=255, blank=True)),
                ('data_volume_size', models.PositiveIntegerField(default=20480, help_text='Data disk size in MiB', validators=[django.core.validators.MinValueValidator(1024)])),
                ('flavor_name', models.CharField(max_length=255, blank=True)),
                ('flavor_disk', models.PositiveIntegerField(default=0, help_text='Flavor disk size in MiB')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InstanceSecurityGroup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('instance', models.ForeignKey(related_name='security_groups', to='openstack.Instance')),
            ],
        ),
        migrations.CreateModel(
            name='OpenStackService',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(verbose_name='organization', to='structure.Customer')),
            ],
            options={
                'verbose_name': 'OpenStack provider',
                'verbose_name_plural': 'OpenStack providers',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OpenStackServiceProjectLink',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('project', models.ForeignKey(to='structure.Project')),
                ('service', models.ForeignKey(to='openstack.OpenStackService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'OpenStack provider project link',
                'verbose_name_plural': 'OpenStack provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SecurityGroup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')])),
                ('backend_id', models.CharField(max_length=128, blank=True)),
                ('service_project_link', models.ForeignKey(related_name='security_groups', to='openstack.OpenStackServiceProjectLink')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SecurityGroupRule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('protocol', models.CharField(blank=True, max_length=4, choices=[('tcp', 'tcp'), ('udp', 'udp'), ('icmp', 'icmp')])),
                ('from_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('to_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('cidr', models.CharField(max_length=32, blank=True)),
                ('backend_id', models.CharField(max_length=128, blank=True)),
                ('security_group', models.ForeignKey(related_name='rules', to='openstack.SecurityGroup')),
            ],
        ),
        migrations.CreateModel(
            name='Tenant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(max_length=150, verbose_name='runtime state', blank=True)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')])),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('internal_network_id', models.CharField(max_length=64, blank=True)),
                ('external_network_id', models.CharField(max_length=64, blank=True)),
                ('availability_zone', models.CharField(help_text='Optional availability group. Will be used for all instances provisioned in this tenant', max_length=100, blank=True)),
                ('user_username', models.CharField(max_length=50, blank=True)),
                ('user_password', models.CharField(max_length=50, blank=True)),
                ('service_project_link', models.ForeignKey(related_name='tenants', on_delete=django.db.models.deletion.PROTECT, to='openstack.OpenStackServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='openstackservice',
            name='projects',
            field=models.ManyToManyField(related_name='openstack_services', through='openstack.OpenStackServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='openstackservice',
            name='settings',
            field=models.ForeignKey(to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='instancesecuritygroup',
            name='security_group',
            field=models.ForeignKey(related_name='instance_groups', to='openstack.SecurityGroup'),
        ),
        migrations.AddField(
            model_name='instance',
            name='service_project_link',
            field=models.ForeignKey(related_name='instances', on_delete=django.db.models.deletion.PROTECT, to='openstack.OpenStackServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='instance',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='service_project_link',
            field=models.ForeignKey(related_name='floating_ips', to='openstack.OpenStackServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='instance',
            field=models.ForeignKey(related_name='backup_schedules', to='openstack.Instance'),
        ),
        migrations.AddField(
            model_name='backup',
            name='backup_schedule',
            field=models.ForeignKey(related_name='backups', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='openstack.BackupSchedule', null=True),
        ),
        migrations.AddField(
            model_name='backup',
            name='instance',
            field=models.ForeignKey(related_name='backups', to='openstack.Instance'),
        ),
        migrations.AlterUniqueTogether(
            name='openstackserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='openstackservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='flavor',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-27 13:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_azure', '0014_networkinterface_ip_address'),
    ]

    operations = [
        migrations.AddField(
            model_name='location',
            name='enabled',
            field=models.BooleanField(default=True, help_text='Indicates whether location is available for resource group.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-06 16:54
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0013_provider_to_customer'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offering',
            name='preferred_language',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def update_vmware_component(apps, schema_editor):
    OfferingComponent = apps.get_model('marketplace', 'OfferingComponent')
    OfferingComponent.objects.filter(
        offering__type='VMware.VirtualMachine',
        type='cpu',
    ).update(
        measured_unit='vCPU'
    )


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0089_make_billing_period_non_nullable')
    ]

    operations = [
        migrations.RunPython(update_vmware_component)
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import taggit.managers
import django_fsm
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0026_floating_ip_resource'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='floatingip',
            name='status',
        ),
    ]
